#ifndef _XRDDPMXMI_H_
#define _XRDDPMXMI_H_
/******************************************************************************/
/*                                                                            */
/*                          X r d D P M X m i . h h                           */
/*                                                                            */
/* DPM handling and various by David Smith, CERN                              */
/* Other parts:                                                               */
/*                                                                            */
/* (c) 2006 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*                            All Rights Reserved                             */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*              DE-AC02-76-SFO0515 with the Department of Energy              */
/******************************************************************************/

#include "XrdOlb/XrdOlbReq.hh"
#include "XrdOlb/XrdOlbXmi.hh"
#include "XrdSys/XrdSysPthread.hh"

#include "dpns_api.h"
#include "serrno.h"

#include <memory>

class XrdSysError;
class XrdInet;
class XrdScheduler;
class XrdOucTrace;
class XrdDPMQueue;

namespace DPMXmi {
class XmiErr;
}

class XrdDPMXmi : XrdOlbXmi
{
public:

       int  Chmod (      XrdOlbReq      *Request,
                   const char           *path,
                         mode_t          mode);

       int  Mkdir (      XrdOlbReq      *Request,
                   const char           *path,
                         mode_t          mode);

       int  Mkpath(      XrdOlbReq      *Request,
                   const char           *path,
                         mode_t          mode);

       // Prep was changed to the signature below in XrdOlbXmi (XrdOlbXmi.hh)
       int  Prep  (const char           *ReqID,
                   const char           *path,
                         int             opts)
                  {return 1;}

       // Older signature of virtual Prep() in XrdOlbXmi
       int  Prep  (      XrdOlbReq      *Request,
                   const char           *path,
                         int             opts)
                  {return notSupported(Request, "prepare", path);}

       int  Rename(      XrdOlbReq      *Request,
                   const char           *oldpath,
                   const char           *newpath);

       int  Remdir(      XrdOlbReq      *Request,
                   const char           *path);

       int  Remove(      XrdOlbReq      *Request,
                   const char           *path);

       int  Select(      XrdOlbReq      *Request,
                   const char           *path,
                         int             opts);

       int  Stat  (      XrdOlbReq      *Request,
                   const char           *path);

       // processing methods called from a XmiDPMReq object
       //
       int      Chmod_process(XrdDPMReq *rp, const bool &destroy);
       int      Mkdir_process(XrdDPMReq *rp, const bool &destroy);
       int      Mkpath_process(XrdDPMReq *rp, const bool &destroy);
       int      Rename_process(XrdDPMReq *rp, const bool &destroy);
       int      Remdir_process(XrdDPMReq *rp, const bool &destroy);
       int      Remove_process(XrdDPMReq *rp, const bool &destroy);
       int      Select_process(XrdDPMReq *rp, const bool &destroy);
       int      Stat_process(XrdDPMReq *rp, const bool &destroy);

       // for getting request specific options from the surl
       //
       int      getExtraSurlOptions(const char *p, const char *key, size_t maxlen,
                                    const char **outp, size_t *outlen) const;
       int      getExtraSurlOptions(const char *Path, XrdDPMReq *rp) const;

       // default lfn2rfn for use by the XmiDPMQueue object
       int      defaultLfn2rfn(const char *Path, char *new_path, size_t s) const;

       void     InitXeq();

       void     MSSPoll();

                XrdDPMXmi(XrdOlbXmiEnv *);

                ~XrdDPMXmi();

       // This XeqMode is the same as the default virtual method in the base class,
       // but we state it explictly in the XrdDPMXmi class
       // 
       void     XeqMode ( unsigned int &isNormal,
                          unsigned int &isDirect)
                    {isNormal = XMI_LOAD | XMI_SPACE; isDirect = 0;}

private:

       // structure with various DPM options: filled from configuration command line args and file
       //
struct dpmoptions_s {
          char           dpns_host[CA_MAXHOSTNAMELEN+1];
          unsigned short dpns_port;
          unsigned short xrd_server_port;
          bool           mkpath_bool;
          time_t         reqput_lifetime;
          char           reqput_ftype;
          char           reqput_stoken[CA_MAXDPMTOKENLEN+1];
          u_signed64     reqput_reqsize;
          time_t         reqget_lifetime;
          char           reqget_ftype;
          char           reqget_stoken[CA_MAXDPMTOKENLEN+1];
       }             Opts;
       //
       // structure for private data for processing chmod, mkdir, mkpath
       //
struct pData1 {
          mode_t         mode;
       };
       //
       // structure for private data for rename
       //
struct pData2 {
          char           opath[CA_MAXPATHLEN+1];
          char           npath[CA_MAXPATHLEN+1];
       };
       //                                                                               
       // structure for private data for select (i.e. open)
       //
struct pData3 {
          int            opts;
          bool           tried_mkpath;
          bool           get,first;
          char           r_token[CA_MAXDPMTOKENLEN+1];
          char           fstatbuf[256];
       };

       int      Select_process_put_first(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error);
       int      Select_process_put(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error);
       int      Select_process_get_first(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error);
       int      Select_process_get(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error);

       void     Init(int When=0, int force=0);
       int      ageSession();
       int      ageSession(const time_t &t);
       int      setupAndUseSession(XrdDPMReq *rp, DPMXmi::XmiErr &err);
       int      setupAndUseSession();
       int      mkp(const char *path, const mode_t &mode) const;
       int      SetDefaultsAndReadConfig(XrdOlbXmiEnv *Env);
const  char    *rcToStr(const int &rc) const;
       int      sendError(XrdOlbReq *reqP, const int &rc, const char *opn, const char *path) const;
       int      sendWait(XrdOlbReq *reqP, const int &rc, const char *opn) const;
       void     sendRedirect(XrdOlbReq *Request, const char *turl, const char *token, const bool &put) const;
       int      notSupported(XrdOlbReq *rp, const char *opn, const char *path) const;

       // Various pointers
       //
static XrdSysError  *eDest;             // -> Error message handler
static XrdInet      *iNet;              // -> Network object
static XrdScheduler *Sched;             // -> Thread scheduler
static XrdOucTrace  *Trace;             // -> Trace handler

       // some constants
       //
static const int     retryTime        =  90; // How long a client waits between retries
static const int     MSSPollTime      =  30; // Poll waiting requests at least every 30 seconds
static const int     maxsessage       =  20; // DPNS session becomes liable to be closed after 20 seconds inactivity

       // Initialization control variable protected by initMutex
       //
       XrdSysMutex   initMutex;
       char          initDone;
       char          initActive;

       // Error buffer for dpns/dpm API
       char          m_apibuf[1024];

       // endpoint, dpns session status and request queue
       //
       char          m_ServerAndPort[CA_MAXHOSTNAMELEN+7];
       bool          m_dpnsSessEst;
       time_t        m_dpnsSessLastChecked;
       XrdDPMQueue  *queue;
};
#endif
